/*
 * Copyright (c) 2008 Nathan Sweet
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
 * IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package net.sourceforge.yamlbeans;

import java.beans.IntrospectionException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import net.sourceforge.yamlbeans.Beans.Property;
import net.sourceforge.yamlbeans.emitter.EmitterConfig;
import net.sourceforge.yamlbeans.scalar.DateSerializer;
import net.sourceforge.yamlbeans.scalar.ScalarSerializer;

/**
 * Stores configuration for reading and writing YAML.
 * @author <a href="mailto:misc@n4te.com">Nathan Sweet</a>
 */
public class YamlConfig {
	/**
	 * Configuration for writing YAML.
	 */
	public final WriteConfig writeConfig = new WriteConfig();

	/**
	 * Configuration for reading YAML.
	 */
	public final ReadConfig readConfig = new ReadConfig();

	final Map<String, String> classNameToTag = new HashMap();
	final Map<String, String> tagToClassName = new HashMap();
	final Map<Class, ScalarSerializer> scalarSerializers = new HashMap();
	final Map<Property, Class> propertyToElementType = new HashMap();

	public YamlConfig () {
		scalarSerializers.put(Date.class, new DateSerializer());
	}

	/**
	 * Allows the specified tag to be used in YAML instead of the full class name.
	 */
	public void setClassTag (String tag, Class type) {
		if (tag == null) throw new IllegalArgumentException("tag cannot be null.");
		if (type == null) throw new IllegalArgumentException("type cannot be null.");
		classNameToTag.put(type.getName(), tag);
		tagToClassName.put(tag, type.getName());
	}

	/**
	 * Adds a serializer for the specified scalar type.
	 */
	public void setScalarSerializer (Class type, ScalarSerializer serializer) {
		if (type == null) throw new IllegalArgumentException("type cannot be null.");
		if (serializer == null) throw new IllegalArgumentException("serializer cannot be null.");
		scalarSerializers.put(type, serializer);
	}

	/**
	 * Sets the type of elements in a Collection property, such as a List.
	 */
	public void setPropertyElementType (Class type, String propertyName, Class elementType) {
		if (type == null) throw new IllegalArgumentException("type cannot be null.");
		if (propertyName == null) throw new IllegalArgumentException("propertyName cannot be null.");
		if (elementType == null) throw new IllegalArgumentException("propertyType cannot be null.");
		Property property;
		try {
			property = Beans.getProperty(type, propertyName);
		} catch (IntrospectionException ex) {
			throw new IllegalArgumentException("The class " + type.getName() + " does not have a property named: " + propertyName,
				ex);
		}
		if (!Collection.class.isAssignableFrom(property.getType()) && !Map.class.isAssignableFrom(property.getType()))
			throw new IllegalArgumentException("The '" + propertyName + "' property on the " + type.getName()
				+ " class must be a Collection or Map: " + property.getType());
		propertyToElementType.put(property, elementType);
	}

	static public class WriteConfig {
		boolean explicitFirstDocument = false;
		boolean writeDefaultValues = false;
		boolean writeRootTags = true;
		boolean autoAnchor = true;
		EmitterConfig emitterConfig = new EmitterConfig();

		WriteConfig () {
			emitterConfig.setUseVerbatimTags(false);
		}

		/**
		 * If true, the first document will have a document start token (---). Default is false.
		 */
		public void setExplicitFirstDocument (boolean explicitFirstDocument) {
			this.explicitFirstDocument = explicitFirstDocument;
		}

		/**
		 * If true, the root of each YAML document will have a tag defining the class that was written. Tags are never written for
		 * primitive types, Strings, {@link ArrayList}, or {@link HashMap}. Default is true.
		 */
		public void setWriteRootTags (boolean writeRootTags) {
			this.writeRootTags = writeRootTags;
		}

		/**
		 * If false, object fields with default values will not be written. A prototype object is created to determine the default
		 * value for each field on the object. Default is false.
		 */
		public void setWriteDefaultValues (boolean writeDefaultValues) {
			this.writeDefaultValues = writeDefaultValues;
		}

		/**
		 * If true, values that are referenced multiple times will use an anchor. This works across YAML documents (ie multiple
		 * calls to {@link YamlWriter#write(Object)}). When true, objects are not actually written until
		 * {@link YamlWriter#clearAnchors()} or {@link YamlWriter#close()} is called. If changing auto anchor to false,
		 * {@link YamlWriter#clearAnchors()} should be called first to output any queued objects. Default is true.
		 */
		public void setAutoAnchor (boolean autoAnchor) {
			this.autoAnchor = autoAnchor;
		}

		/**
		 * Sets the YAML version to output. Default is 1.1.
		 */
		public void setVersion (Version version) {
			emitterConfig.setVersion(version);
		}

		/**
		 * If true, the YAML output will be canonical. Default is false.
		 */
		public void setCanonical (boolean canonical) {
			emitterConfig.setCanonical(canonical);
		}

		/**
		 * Sets the number of spaces to indent. Default is 3.
		 */
		public void setIndentSize (int indentSize) {
			emitterConfig.setIndentSize(indentSize);
		}

		/**
		 * Sets the column at which values will attempt to wrap. Default is 100.
		 */
		public void setWrapColumn (int wrapColumn) {
			emitterConfig.setWrapColumn(wrapColumn);
		}

		/**
		 * If false, tags will never be surrounded by angle brackets (eg, "!<java.util.LinkedList>"). Default is false.
		 */
		public void setUseVerbatimTags (boolean useVerbatimTags) {
			emitterConfig.setUseVerbatimTags(useVerbatimTags);
		}
	}

	static public class ReadConfig {
		Version defaultVersion = new Version(1, 1);

		ReadConfig () {
		}

		/**
		 * Sets the default YAML version to expect if a YAML document does not explicitly specify a version. Default is 1.1.
		 */
		public void setDefaultVersion (Version defaultVersion) {
			if (defaultVersion == null) throw new IllegalArgumentException("defaultVersion cannot be null.");
			this.defaultVersion = defaultVersion;
		}
	}
}
